/* Typedefs */

typedef struct dtd *Dtd;

typedef struct entity *Entity;

typedef struct element_definition *ElementDefinition;

typedef struct attribute_definition *AttributeDefinition;
AttributeDefinition NextAttributeDefinition(ElementDefinition element,
					    AttributeDefinition previous);

typedef struct notation_definition *NotationDefinition;

/* DTDs */

struct dtd {
    const Char *name;		/* The doctype name */
    Entity entities;
    Entity parameter_entities;
    Entity predefined_entities;
#ifdef FOR_NSL
    NSL_Doctype_I *doctype;
#else
    ElementDefinition elements;
#endif
    NotationDefinition notations;
};

/* Entities */

enum entity_type {ET_external, ET_internal};
typedef enum entity_type EntityType;

enum markup_language {ML_xml, ML_nsl, ML_unspecified};
typedef enum markup_language MarkupLanguage;

struct entity {
    /* All entities */

    const Char *name;		/* The name in the entity declaration */
    EntityType type;		/* ET_external or ET_internal */
    const char8 *base_url;	/* If different from expected */
    struct entity *next;	/* For chaining a document's entity defns */
    CharacterEncoding encoding;	/* The character encoding of the entity */
    Entity parent;		/* The entity in which it is defined */
    const char8 *url;		/* URL of entity */

    /* Internal entities */

    const Char *text;		/* Text of the entity */
    int line_offset;		/* Line offset of definition */
    int line1_char_offset;	/* Char offset on first line */ 
    int matches_parent_text;	/* False if might contain expanded PEs */

    /* External entities */

    const char8 *systemid;	/* Declared public ID */
    const char8 *publicid;	/* Declared public ID */
    NotationDefinition notation; /* Binary entity's declared notation */
    MarkupLanguage ml_decl;	/* XML, NSL or not specified */
    const char8 *version_decl;	/* XML declarations found in entity, if any  */
    const char8 *encoding_decl;
    const char8 *standalone_decl;
    const char8 *ddb_filename;	/* filename in NSL declaration */
};

/* Elements */

enum content_type {
    /* NB this must match NSL's ctVals */
    CT_mixed, CT_any, CT_bogus1, CT_bogus2, CT_empty, CT_element, CT_enum_count
};
typedef enum content_type ContentType;

extern const char8 *ContentTypeName[CT_enum_count];

struct element_definition {
    const Char *name;		/* The element name */
    int namelen;
    int tentative;
#ifdef FOR_NSL
    NSL_Doctype_I *doctype;
    NSL_ElementSummary_I *elsum;
#else
    ContentType type;		/* The declared content */
    Char *content;		/* Element content */
    AttributeDefinition attributes;
    struct element_definition *next;
#endif
};

/* Attributes */

enum default_type {
    /* NB this must match NSL's NSL_ADefType */
    DT_required, DT_bogus1, DT_implied, 
    DT_bogus2, DT_none, DT_fixed, DT_enum_count
};
typedef enum default_type DefaultType;

extern const char8 *DefaultTypeName[DT_enum_count];

enum attribute_type {
    /* NB this must match NSL's NSL_Attr_Dec_Value */
    AT_cdata, AT_bogus1, AT_bogus2, AT_nmtoken, AT_bogus3, AT_entity,
    AT_idref, AT_bogus4, AT_bogus5, AT_nmtokens, AT_bogus6, AT_entities,
    AT_idrefs, AT_id, AT_notation, AT_enumeration, AT_enum_count
};
typedef enum attribute_type AttributeType;

extern const char8 *AttributeTypeName[AT_enum_count];

struct attribute_definition {
#ifdef FOR_NSL
    /* NB this must match NSL's AttributeSummary */
    /* We never really have one of these structures; only an AttributeSummary
       cast to this type.  We need to be able to access the type, so that
       we can normalise if appropriate.  We need to be able to refer to
       the default_type and default_value, but these don't need to work
       since we will never have ReturnDefaultedAttributes true in NSL. */
    int a, b, c;
    short d;
    char type, default_type;
    /* This had better never be accessed! */
    Char *default_value;
#else
    const Char *name;		/* The attribute name */
    int namelen;
    AttributeType type;		/* The declared type */
    Char **allowed_values;	/* List of allowed values, argv style */
    DefaultType default_type;	/* The type of the declared default */
    const Char *default_value;	/* The declared default value */
    struct attribute_definition *next;
#endif
};

/* Notations */

struct notation_definition {
    const Char *name;		/* The notation name */
    int tentative;
    const char8 *systemid;	/* System identifier */
    const char8 *publicid;	/* Public identifier */
    struct notation_definition *next;
};

/* Public functions */

Dtd NewDtd(void);

Entity NewExternalEntityN(const Char *name, int namelen,
			  const char8 *publicid, const char8 *systemid,
			  NotationDefinition notation,
			  Entity parent);
Entity NewInternalEntityN(const Char *name, int namelen,
			  const Char *text, Entity parent,
			  int line_offset, int line1_char_offset, 
			  int matches_parent_text);

const char8 *EntityURL(Entity e);
const char8 *EntityDescription(Entity e);
void EntitySetBaseURL(Entity e, const char8 *url);
const char8 *EntityBaseURL(Entity e);

Entity DefineEntity(Dtd dtd, Entity entity, int pe);
Entity FindEntityN(Dtd dtd, const Char *name, int namelen, int pe);

#define NewExternalEntity(name, pub, sys, not, parent) \
    NewExternalEntityN(name, name ? Strlen(name) : 0, pub, sys, not, parent)
#define NewInternalEntity(name, test, parent, l, l1, mat) \
    NewInternalEntityN(name, name ? Strlen(name) : 0, test, parent, l, l1, mat)
#define FindEntity(dtd, name, pe) FindEntityN(dtd, name, Strlen(name), pe)

ElementDefinition DefineElementN(Dtd dtd, const Char *name, int namelen,
				 ContentType type, Char *content);
ElementDefinition TentativelyDefineElementN(Dtd dtd, 
					    const Char *name, int namelen);
ElementDefinition RedefineElement(ElementDefinition e, ContentType type,
				  Char *content);
ElementDefinition FindElementN(Dtd dtd, const Char *name, int namelen);

#define DefineElement(dtd, name, type, content) \
    DefineElementN(dtd, name, Strlen(name), type, content)
#define TentativelyDefineElement(dtd, name) \
    TentativelyDefineElementN(dtd, name, Strlen(name))
#define FindElement(dtd, name) FindElementN(dtd, name, Strlen(name))

AttributeDefinition DefineAttributeN(ElementDefinition element,
				     const Char *name, int namelen,
				     AttributeType type, Char **allowed_values,
				     DefaultType default_type, 
				     const Char *default_value);
AttributeDefinition FindAttributeN(ElementDefinition element,
				   const Char *name, int namelen);

#define DefineAttribute(element, name, type, all, dt, dv) \
    DefineAttributeN(element, name, Strlen(name), type, all, dt, dv)
#define FindAttribute(element, name) \
    FindAttributeN(element, name, Strlen(name))

NotationDefinition DefineNotationN(Dtd dtd, const Char *name, int namelen,
				 const char8 *publicid, const char8 *systemid);
NotationDefinition TentativelyDefineNotationN(Dtd dtd,
					      const Char *name, int namelen);
NotationDefinition RedefineNotation(NotationDefinition n,
				 const char8 *publicid, const char8 *systemid);
NotationDefinition FindNotationN(Dtd dtd, const Char *name, int namelen);

#define DefineNotation(dtd, name, pub, sys) \
    DefineNotationN(dtd, name, Strlen(name), pub, sys)
#define TentativelyDefineNotation(dtd, name) \
    TentativelyDefineNotationN(dtd, name, Strlen(name))
#define FindNotation(dtd, name) FindNotationN(dtd, name, Strlen(name))

